
-------------------------------------------------------------------------------
--! @file       Controller.vhd
--! @author     William Diehl
--! @brief      SCREAM Controller
--! @date       4 Jul 2016
-------------------------------------------------------------------------------

-- One round per clock cycle

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;

ENTITY Controller IS


PORT(
	clk: in std_logic;
	reset: in std_logic;
	en : out std_logic;
	load_npub : out std_logic;
	load_key : out std_logic;
	auth_sum_init : out std_logic;
	tag_valid : in std_logic;

	key_ready : out std_logic;
	key_valid : in std_logic;
	key_update : in std_logic;
	bdi_ready : out std_logic;
	bdi_valid : in std_logic;
	bdi_type  : in  std_logic_vector(3                -1 downto 0);
	bdi_size : in  std_logic_vector(4 downto 0);
	decrypt_init : out std_logic;
	decrypt	: in std_logic;
	bdi_eot : in std_logic;
	bdi_eoi : in std_logic;

	block_counter_init : out std_logic;
	en_block_counter : out std_logic;
	load_sum : out std_logic;
	load_auth : out std_logic;
	load_tag : out std_logic;
	tag_mode : out std_logic;
	null_length : out std_logic;
	
	Ek_sel : out std_logic_vector(1 downto 0);
	bdo_sel : out std_logic_vector(1 downto 0);
	sel_data : out std_logic;
	Ek_start : out std_logic;
	Ek_done : in std_logic;
	inv : out std_logic;

	bdo_valid : out std_logic;
	bdo_ready : in std_logic;
	msg_auth_valid : out std_logic;
	msg_auth_done : out std_logic

	);
	
END Controller;

ARCHITECTURE behavioral OF Controller IS

SIGNAL PS, NS : STD_LOGIC_VECTOR(7 DOWNTO 0);
signal null_length_init, load_null_length : std_logic;

BEGIN

sync_process: PROCESS(clk, reset)
BEGIN

-- Requires 1 asynchronous reset to put in State 0

IF (reset = '1') THEN
	PS <= x"00";
	
-- All state transitions occur on rising edge
-- All transitions are disabled if Read/Write Enable is low

ELSIF (rising_edge(clk)) 
	 THEN PS <= NS;
	     
END IF;
	
END Process;

Public_process: PROCESS(PS, bdi_eot, bdi_eoi, bdi_valid, key_update, key_valid, bdi_type, Ek_done,
								decrypt, tag_valid, bdo_ready, bdi_size)

 BEGIN
 
	-- default values
	decrypt_init <= '0';
	null_length_init <= '0';
	load_null_length <= '0';
	msg_auth_valid <= '0';
	msg_auth_done <= '0';
	auth_sum_init <= '0';
	bdo_valid <= '0';
	bdi_ready <= '0';
	key_ready <= '0';
	load_npub <= '0';
	load_key <= '0';
	en <= '0';
   block_counter_init <= '0';
	en_block_counter <= '0';
	load_sum <= '0';
	load_auth <= '0';
	load_tag <= '0';
	tag_mode <= '0';
	--Ek_sel <= "00";	--! Ek_sel is a latch
	--bdo_sel <= "00";--! bdo_sel is a latch
	Ek_start <= '0';
	
	 case PS IS
		 
		 WHEN x"00" => -- s_init
-- State init_state
-- Initialization after asynchronous reset
		null_length_init <= '1';
		block_counter_init <= '1';
		auth_sum_init <= '1';
		inv <= '0'; -- latch
		NS <= x"01";
				 
		WHEN x"01" => -- s_key_update
	 -- Key check
		decrypt_init <= '1';
		if (bdi_valid = '1') then
			if (key_update = '1') then
				 if (key_valid = '1') then
						key_ready <= '1';
						load_key <= '1';
						NS <= x"02"; 
				 else
					   NS <= x"01";
				 end if;
			else 
				 NS <= x"02";
			end if;
		else NS <= x"01";
		
		end if;				
				
	 WHEN x"02" => -- s_load_npub
	     	load_npub <= '1';
				if (bdi_eot = '1' and bdi_eoi = '1') then
					load_null_length <= '1';
					NS <= x"10"; -- no AD or message - generate tag
	 				
				else
					bdi_ready <= '1'; -- get the first segment and see what it is 
					NS <= x"03";
			   end if;
		 
		 WHEN x"03" => -- s_start_AD
		 
		if (bdi_valid = '1') then
			if (bdi_type(1) = '1') then -- first segment is not AD - go to message
				en <= '1';
				Ek_start <= '1';
				if (decrypt = '0') then
					load_sum <= '1';
				end if;
				if (bdi_eot = '1') then
					if (bdi_size(4) = '1') then -- last block is full
						inv <= decrypt; -- can be encrypt or decrypt
						Ek_sel <= "00";
					else
					   inv <= '0'; -- if last block and partial block only encrypt
						Ek_sel <= "10";
					end if;
					NS <= x"0d";  
				else
					inv <= decrypt; -- can be encrypt or decrypt
					Ek_sel <= "00";
					NS <= x"09"; 
				end if;
			else
				en <= '1';
				inv <= '0'; -- AD blocks are only encrypt
				Ek_start <= '1';
				Ek_sel <= "00";
				NS <= x"04"; 
			end if;
		else
			NS <= x"03";
		end if;
		
		WHEN x"04" => -- s_delay1
		-- this state necessary to hold bdi_type and bdi_size constant for 
		-- tweak key generation
		
		en <= '1';
		NS <= x"05"; 
		
		WHEN x"05" => -- s_get_next_bdi1 
		
		en <= '1';
		if (bdi_eot = '1') then
				if (bdi_eoi = '1') then -- null M
					load_null_length <= '1';
					NS <= x"0f"; -- finish AD and go to tag
				else -- bdi_eoi = '0'
				   bdi_ready <= '1';
					block_counter_init <= '1';
					NS <= x"07"; -- finish AD
				end if;
		else -- bdi_eot = '0'
				bdi_ready <= '1';
				en_block_counter <= '1';
				NS <= x"06"; -- process AD
		end if;
		
		WHEN x"06" => -- s_finish_AD (not last block)
		
		if (Ek_done = '1') then
			load_auth <= '1';
			NS <= x"03";
		else 
			en <= '1';
			NS <= x"06"; 
		end if;
		
		WHEN x"07" => -- s_finish_AD (last block)
		
		if (Ek_done = '1') then
			load_auth <= '1';
			NS <= x"08"; 
		else
			en <= '1';
			NS <= x"07"; 
		end if;
		
		WHEN x"08" => -- s_start_M
		
		if (bdi_valid = '1') then
			en <= '1';
			Ek_start <= '1';
			if (decrypt = '0') then
				load_sum <= '1';
			end if;
			if (bdi_eot = '1') then
			
					if (bdi_size(4) = '1') then -- last block is full
						inv <= decrypt; -- can be encrypt or decrypt
						Ek_sel <= "00";
					else
					   inv <= '0'; -- if last block and partial block only encrypt
						Ek_sel <= "10";
					end if;
				
				NS <= x"0d";  
			else
				inv <= decrypt; -- can be encrypt or decrypt
				Ek_sel <= "00";
				NS <= x"09"; 
			end if;
		
		else
			NS <= x"08"; 
		end if;
			
		WHEN x"09" => -- s_delay2
		-- this state necessary to hold bdi_type and bdi_size constant for 
		-- tweak key generation
		en <= '1';
		NS <= x"0a"; 
		
		WHEN x"0a" => -- s_get_next_bdi2
		
		en <= '1';
		bdi_ready <= '1';
		en_block_counter <= '1';
		
		NS <= x"0b"; 
		
		WHEN x"0b" => -- s_process_M
		
		if (Ek_done = '1') then
			if (decrypt = '1') then
				load_sum <= '1';
			end if;
			if (bdo_ready = '1') then
				bdo_sel <= "00";
				bdo_valid <= '1';
				NS <= x"08"; 
			else 
				NS <= x"0c"; -- wait for bdo valid
			end if;
		else
			en <= '1';
			NS <= x"0b"; 
		end if;
		
		WHEN x"0c" => -- s_wait_for_bdo_valid1
		
		if (bdo_ready = '1') then
			bdo_sel <= "00";
			bdo_valid <= '1';
			NS <= x"08"; 
		else
			NS <= x"0c"; 
		end if;
		
		WHEN x"0d" => -- s_finish_final_M 
		
		if (Ek_done = '1') then
			if (decrypt = '1') then
				load_sum <= '1';
			end if;
			if (bdo_ready = '1') then
				if (bdi_size(4) = '1') then -- full block
					en_block_counter <= '1';
					bdo_sel <= "00";
				else
					bdo_sel <= "10";
				end if;
				bdo_valid <= '1';
				NS <= x"10"; 
			else 
				NS <= x"0e"; -- wait for bdo valid
			end if;
		else
			en <= '1';
			NS <= x"0d"; 
		end if;
		
		WHEN x"0e" => -- s_wait_for_bdo_valid2
		
		if (bdo_ready = '1') then
		   if (bdi_size(4) = '1') then -- full block
					en_block_counter <= '1';
					bdo_sel <= "00";
				else
					bdo_sel <= "10";
				end if;
			bdo_valid <= '1';
			NS <= x"10"; 
		else
			NS <= x"0c"; 
		end if;
				
		WHEN x"0f" => -- s_finish_AD and go to tag
		
		if (Ek_done = '1') then
			load_auth <= '1';
			NS <= x"10"; 
		else
			en <= '1';
			NS <= x"0f"; 
		end if;
		
		WHEN x"10" => -- s_generate_tag
		
		en <= '1';
		Ek_sel <= "11";
		inv <= '0'; -- tag generation is only encrypt
		Ek_start <= '1';
		tag_mode <= '1';
		NS <= x"11"; 
		
		WHEN x"11" => -- s_get_next_bdi2
		
		tag_mode <= '1';
		en <= '1';
		NS <= x"12"; 
		
		WHEN x"12" => -- s_delay3
		-- this state necessary to hold bdi_type and bdi_size constant for 
		-- tweak key generation
		tag_mode <= '1';
		bdi_ready <= '1';
		en <= '1';
		NS <= x"13"; 
		
		WHEN x"13" => -- s_finish_tag
		
		if (Ek_done = '1') then
			if (decrypt = '1') then
				load_tag <= '1';
				NS <= x"14"; -- verify tag
			else -- decrypt = '0'
				if (bdo_ready = '1') then
					bdo_sel <= "11";
					bdo_valid <= '1';
					NS <= x"00"; -- done
				else 
					NS <= x"15"; -- wait for bdo ready
				end if;
			end if;
		else
			en <= '1';
			tag_mode <= '1';
			NS <= x"13"; 
		end if;
		
		WHEN x"14" => -- s_verify_tag
		
		if (bdi_valid = '1') then
			msg_auth_done <= '1';
			if (tag_valid = '1') then
				msg_auth_valid <= '1';
			end if;
			bdi_ready <= '1';
			NS <= x"00";
		else
			NS <= x"14"; 
		end if;
		
		WHEN x"15" => -- s_wait_for_bdo_ready3
					
		if (bdo_ready = '1') then
				bdo_sel <= "11";
				bdo_valid <= '1';
				NS <= x"00"; -- done
		else 
				NS <= x"15"; -- wait for bdo ready
		end if;
		
		WHEN OTHERS => 
		
		NS <= x"00";
			  
	  END case; 
END Process;

store_null_length: entity work.reg1(behavioral)
	port map(
		d => '1',
		--reset => rst,
		clk => clk,
		init => null_length_init,
		en => load_null_length,
		q => null_length);

END behavioral;  