
-------------------------------------------------------------------------------
--! @file       Datapath.vhd
--! @author     William Diehl
--! @brief      SCREAM Datapath
--! @date       4 Jul 2016
-------------------------------------------------------------------------------

-- One round per clock cycle

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.all;

ENTITY Datapath IS
	GENERIC(
	--! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
		  G_MAX_LEN		   : integer := 32
		  );
	PORT(
	 clk: IN STD_LOGIC;
	 --rst: IN STD_LOGIC;
	en : in std_logic;
	 
	 --rand_seed : IN std_logic_vector(G_DBLK_SIZE - 1 downto 0); -- used only in masked version
	bdi : in std_logic_vector(G_DBLK_SIZE - 1 downto 0);  
	bdo : out std_logic_vector(G_DBLK_SIZE - 1 downto 0);
	key_in : in std_logic_vector(G_KEY_SIZE - 1 downto 0);

	bdi_type : in std_logic_vector(3 - 1 downto 0);
	bdi_size : std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
	bdi_eot : in std_logic;
	bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
   
	decrypt : in std_logic;
	decrypt_init : in std_logic;
	decrypt_reg_out : out std_logic;
	load_key : in std_logic;	
	load_npub : in std_logic;
	block_counter_init : in std_logic;
	en_block_counter : in std_logic;
	auth_sum_init : in std_logic;
	load_sum : in std_logic;
	load_auth : in std_logic;
	load_tag : in std_logic;
	tag_valid : out std_logic;
	tag_mode : in std_logic;
	null_length : in std_logic;

   Ek_inv : in std_logic;
	Ek_sel : in std_logic_vector(1 downto 0);
	bdo_sel : in std_logic_vector(1 downto 0);
	Ek_start : in std_logic;
	Ek_done : out std_logic

);
	 
END Datapath;

ARCHITECTURE structural of Datapath IS

CONSTANT NS: INTEGER:= 10; -- Number of steps (currently works from 8 to 12 steps)

signal decrypt_reg : std_logic;
signal key : std_logic_vector(G_KEY_SIZE - 1 downto 0);
signal npub : std_logic_vector(87 downto 0);
signal next_block_counter, block_counter : std_logic_vector(G_MAX_LEN - 4 - 1 downto 0);
signal tweak0 : std_logic_vector(G_DBLK_SIZE - 1 downto 0);
signal p_last : std_logic_vector(G_DBLK_SIZE - 1 downto 0);
signal Ek_din, Ek_dout :std_logic_vector(G_DBLK_SIZE - 1 downto 0); 
signal next_sum, sum, sum_in : std_logic_vector(G_DBLK_SIZE - 1 downto 0); 
signal output_masking : std_logic_vector(G_DBLK_SIZE - 1 downto 0); 
signal Ek_dout_truncated, Ek_dout_trunc_xor_p : std_logic_vector(G_DBLK_SIZE - 1 downto 0); 
signal next_auth, auth : std_logic_vector(G_TAG_SIZE - 1 downto 0);
signal next_tag, tag : std_logic_vector(G_TAG_SIZE - 1 downto 0);
signal sum_sel : std_logic_vector(1 downto 0);
signal block_full, block_full_reg, block_full_in : std_logic;

--signal rand_num : std_logic_vector(127 downto 0); --used only in masked version

BEGIN

	--rand_num <= rand_seed xor w; -- random number for masking used only in masked version
  				
	decrypt_reg_out <= decrypt_reg;
	
	store_decrypt: entity work.reg1(behavioral)
	port map(
		d => decrypt,
		--reset => rst,
		clk => clk,
		init => '0',
		en => decrypt_init,
		q => decrypt_reg);

	store_key: entity work.regn_en(behavioral)
		generic map(n=>G_KEY_SIZE)
		port map(
			d => key_in,
			--reset => rst,
			clk => clk,
			init => '0',
			en => load_key,
			q => key);

	store_npub: entity work.regn_en(behavioral)
	generic map(n=>88)
	port map(
		d => bdi(G_DBLK_SIZE - 1 downto G_DBLK_SIZE - 88),
		--reset => rst,
		clk => clk,
		init => '0',
		en => load_npub,
		q => npub);

 next_block_counter <= block_counter + 1;

	store_block_counter: entity work.regn_en(behavioral)
	generic map(n=>(G_MAX_LEN-4)) -- 32 or 11
	port map(
		d => next_block_counter,
		--reset => rst,
		clk => clk,
		init => block_counter_init,
		en => en_block_counter,
		q => block_counter);

   	get_Tweak_Input: entity work.Tweak_Input(dataflow)
            GENERIC MAP(NS => NS,
							   G_DBLK_SIZE => G_DBLK_SIZE,
								G_LBS_BYTES => G_LBS_BYTES,
								G_MAX_LEN =>   G_MAX_LEN				
				)
		PORT MAP(
		    decrypt => Ek_inv,
		    bdi_type => bdi_type,
			 tag_mode => tag_mode,
			 null_length => null_length,
          bdi_size => bdi_size,
          bdi_eot => bdi_eot,
		    CN => block_counter,	
          N => npub,
          T => tweak0
);

--! values of block_full must be buffered and retained throughout the calculation

	store_block_full: entity work.reg1(behavioral)
	port map(
		d => block_full_in,
		--reset => rst,
		clk => clk,
		init => '0',
		en => Ek_start,
		q => block_full_reg);	
	
	block_full_in <= '1' when (bdi_size(G_LBS_BYTES) = '1') else '0';
	block_full <= block_full_in when (Ek_start = '1') else block_full_reg;
		
	p_last <=  x"000000000000000000000000000000" & bdi_size & "000";
	
	with Ek_sel select
    		Ek_din <= bdi when "00",			
		     bdi when "01",			
			  p_last when "10",			
	        sum when "11", 
	        bdi when others;

   	compute_Ek: entity work.Ek(structural)
           GENERIC MAP(NS => NS)
       	   PORT MAP(
               CLK => clk,
               --RESET => rst,
					rho_init => block_counter_init,
               INIT => Ek_start,
					EN => en,
					T0_rev => tweak0,
               K_rev => key,
               end_of_block => Ek_done,
	       --x_prime => rand_num, --used only in masked version
               X_IN_rev => Ek_din,
               X_OUT_rev => Ek_dout,
					inv => Ek_inv
		);
		
-- Prepare and store checksum
	sum_sel(1) <= decrypt_reg;
	sum_sel(0) <= block_full; 

	with sum_sel select
    		sum_in <= bdi when "00",			
		       bdi when "01",			
			    Ek_dout_trunc_xor_p when "10",			
	          Ek_dout when "11", 
	          bdi when others;
        
	next_sum <= sum xor sum_in;

-- Trunc(E(T,|Pi+1|)

	Ek_dout_truncated <= Ek_dout and output_masking;
	
	g1: for i in 0 to 15 generate
		output_masking((15 - i) * 8 + 7 downto (15 - i) * 8) <= bdi_valid_bytes(15-i) & 
		bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) &
		bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i);
	end generate g1;
		
	store_sum: entity work.regn_en(behavioral)
                 GENERIC MAP(N=>128)
                 PORT MAP(
                       clk => clk,
                       --RESET => rst,
                       INIT => auth_sum_init,
                       EN => load_sum,
                       D => next_sum,
                       Q => sum);

-- Prepare and store auth

	next_auth <= auth xor Ek_dout;

	store_auth: entity work.regn_en(behavioral)
                 GENERIC MAP(N=>128)
                 PORT MAP(
                       CLK => clk,
                       --RESET => Rst,
                       INIT => auth_sum_init,
                       EN => load_auth,
                       D => next_auth,
                       Q => auth);

-- Select bdo for output

	Ek_dout_trunc_xor_p <= Ek_dout_truncated xor bdi;

	with bdo_sel select
    		bdo <= Ek_dout when "00",
		       Ek_dout when "01",	
		       Ek_dout_trunc_xor_p when "10",
		       next_tag when "11",
		       Ek_dout when others;

-- generate and verify tag
	next_tag <= Ek_dout xor auth;

	store_tag: entity work.regn_en(behavioral)
	generic map(n=>128) 
	port map(
		d => next_tag,
		--reset => rst,
		clk => clk,
		init => '0',
		en => load_tag,
		q => tag);

	tag_valid <= '1' when (tag = bdi) else '0';

END structural;