
-------------------------------------------------------------------------------
--! @file       Ek_Round.vhd
--! @author     William Diehl
--! @brief      SCREAM Ek_Round
--! @date       4 Jul 2016
-------------------------------------------------------------------------------

-- Ek Round
-- Ek Round implemented with one round per clock cycle

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;

ENTITY Ek_Round IS
	PORT(
	 
	 X_IN : IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 X_OUT : OUT STD_LOGIC_VECTOR(127 DOWNTO 0);
	 SIGMA  : IN STD_LOGIC_VECTOR(3 DOWNTO 0);
	 RHO : IN STD_LOGIC; -- referenced for encryption; decryption is not(rho)
	 T: IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 Next_T : OUT STD_LOGIC_VECTOR(127 DOWNTO 0);
	 K: IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 inv : in std_logic); -- inv=0 encrypt; inv=1 decrypt

END Ek_Round;

ARCHITECTURE structural of Ek_Round IS

signal s_in, s_out : std_logic_vector(127 downto 0);
SIGNAL X0e : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL A0e, A1e, B0e, B1e : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL A0d, A1d, B0d, B1d : STD_LOGIC_VECTOR(127 DOWNTO 0);

SIGNAL C0e, C1e : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL C1d : STD_LOGIC_VECTOR(15 DOWNTO 0);
 
SIGNAL RC0, RC0_table, RC1, RC1_table : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL X_before_TK, TK, TSIGMA, TSIGMAe, TSIGMAd: STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL Txor : std_logic_vector(63 downto 0);

BEGIN


-- Global 

compute_RC_0: entity work.RC0_LUT(dataflow)
	  PORT MAP(
		sigma => sigma,
		RC => RC0_table
		);

compute_RC_1: entity work.RC1_LUT(dataflow)
	  PORT MAP(
		sigma => sigma,
		RC => RC1_table
		);

Txor <= T(127 DOWNTO 64) XOR T(63 DOWNTO 0);

-- align round constants

RC0 <= RC0_table when (rho='0') else RC1_table;
RC1 <= RC1_table when (rho='0') else RC0_table;


-- Encrypt sequence

	s_in <= x_in when (inv='0') else B1d;

	compute_S_Box: entity work.Sbox_bidir(structural)
		PORT MAP(
		X_IN => s_in,
      inv => inv, 
		X_OUT => s_out);	

-- Round Constant addition after S Box in encryption sequence

     A0e <= s_out;
			
     C0e <= A0e(127 DOWNTO 112) XOR RC0;
-- RC addition to Most significant word modulo 65536

   B0e<= C0e & A0e(111 DOWNTO 0);
-- Add Round Constant to Most significant word of status register

        compute_L_Box: entity work.L_Box_128_alt(structural)
	PORT MAP(
		
		x => B0e,
	 	w => X0e);

-- Decrypt sequence


-- Inv_L_Box_128 is 8-bit Inv LBox
-- Inv_L_Box_128_alt is 16-bit Inv LBox

   compute_Inv_L_Box_1: entity work.Inv_L_Box_128_alt(structural)
	PORT MAP(
	      x => X_IN,
	      w => A1d);
	
-- Round Constant addition after Inv L Box in decryption sequence

     C1d <= A1d(127 DOWNTO 112) XOR RC1;
-- RC addition to most signficant word modulo 65536

     B1d <= C1d & A1d(111 DOWNTO 0);

-- Add Round Constant to Most significant word of status register
	

-- Common termination sequence


     X_before_TK <= X0e when (inv = '0') else s_out;

-- compute next tweak

     TSIGMAe <= Txor & T(127 DOWNTO 64); 
     TSIGMAd <= T(63 DOWNTO 0) & Txor;

     TSIGMA <= TSIGMAe when (inv='0') else TSIGMAd;

     TK <= TSIGMA XOR K;
-- Compute Step Tweak Key

     X_OUT <= (X_before_TK XOR TK) when (rho = '1') else X_before_TK;

-- Input XOR with Tweak Key

    Next_T <= TSIGMA;

END structural;
