-------------------------------------------------------------------------------
--! @file       Ek_bidir.vhd
--! @author     William Diehl
--! @brief      SCREAM Ek_bidir
--! @date       29 Jun 2016
-------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.all;

ENTITY EK_bidir IS
	GENERIC(
	NS : INTEGER:=12 -- Number of steps
	);
	PORT(
	 CLK: IN STD_LOGIC;
	 --RESET: IN STD_LOGIC;
	 INIT: IN STD_LOGIC;
	 EN : IN STD_LOGIC;
	 T0_rev : IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 K_rev: IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 end_of_block : OUT STD_LOGIC;
	 --x_prime : in std_logic_vector(127 downto 0); -- used only in masked version
	 X_IN_rev : IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 X_OUT_rev : OUT STD_LOGIC_VECTOR(127 DOWNTO 0);
	 inv : in std_logic -- inv=0 is encrypt; inv = 1 is decrypt
);
	 
END EK_bidir;

ARCHITECTURE structural of EK_bidir IS

SIGNAL X_IN : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL X_OUT : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL K : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL T0 : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL T_out: STD_LOGIC_VECTOR(127 DOWNTO 0);

SIGNAL X, Next_X, X_to_Ek : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL X_IN_Tweak : STD_LOGIC_VECTOR(127 DOWNTO 0);

SIGNAL T, Next_T, T_to_Ek : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL SIGMA, Next_SIGMA, NEXT_SIGMA_inc, NEXT_SIGMA_dec: STD_LOGIC_VECTOR(3 DOWNTO 0);
signal next_sigma_start, next_sigma_run : std_logic_vector(3 downto 0);
SIGNAL Sigma_to_Ek : std_logic_vector(3 downto 0);
SIGNAL SEL, SEL_enc, SEL_dec, end_of_block_enc, end_of_block_dec: STD_LOGIC;

--! used only in masked version
--signal next_x_prime, x_prime_to_Ek, x_prime_reg: std_logic_vector(127 downto 0);
--signal x_in_tweak_prime, rx_in_tweak_prime, rk : std_logic_vector(127 downto 0);

BEGIN

NEXT_SIGMA_inc <= SIGMA + 1;
NEXT_SIGMA_dec <= SIGMA - 1;

next_sigma_run <= NEXT_SIGMA_inc when (inv='0') else NEXT_SIGMA_dec;
next_sigma_start <= x"0" when (inv = '0') else std_logic_vector(to_unsigned(NS,4));
NEXT_SIGMA <= next_sigma_start when (init = '1') else next_sigma_run;

SEL_enc <= '1' when (SIGMA = 0) else '0';
SEL_dec <= '1' when (SIGMA= NS) else '0';
SEL <= SEL_enc when (inv = '0') else SEL_dec;

end_of_block_enc <= '1' when ((SIGMA = NS) and init = '0') else '0';
end_of_block_dec <= '1' when ((SIGMA = 0) and init = '0') else '0';

end_of_block <= end_of_block_enc when (inv = '0') else end_of_block_dec;

-- mask initial tweak key - the secret key is the sensitive variable
--! used only in masked version
--rk <= k xor x_prime;
--x_in_tweak_prime <= x_in xor x_prime;
--rx_in_tweak_prime <= rk xor t0;
--x_in_tweak <= x_in_tweak_prime xor rx_in_tweak_prime;

X_IN_Tweak <= X_IN XOR K XOR T0;
T_to_Ek <= T0 WHEN (SEL='1') ELSE T;
X_to_Ek <= X_IN_Tweak WHEN (SEL='1') ELSE X;

Sigma_to_Ek <= SIGMA when (inv='0') else NEXT_SIGMA;

--! used only in masked version
--x_prime_to_Ek <= x_prime when (SEL = '1') else x_prime_reg;
--next_x_prime <= x_prime xor Next_X;

-- select Ek_Step_bidir_mod for shared S-Boxes

   step: entity work.Ek_Step_bidir(structural) 
		
   	PORT MAP(
		 	
			T => T_to_Ek,
			Next_T => Next_T,
			SIGMA => Sigma_to_Ek,
			--x_prime => x_prime_to_Ek, -- random masking variable used only in masked version
			X_IN => X_to_Ek,
			X_OUT => Next_X,
			K => K,
			inv => inv
			);

   store_X: entity work.regn(behavioral)
		 GENERIC MAP(N=>128)
		 PORT MAP(
			CLK => CLK,
			--RESET => RESET,
			EN => EN,
			D => Next_X,
			Q => X);
						
--! used only in masked version
	--store_X_prime: entity work.regn(behavioral)
	--	 GENERIC MAP(N=>128)
		-- PORT MAP(
		--	CLK => CLK,
		--	RESET => RESET,
		--	EN => EN,
		--	D => next_x_prime,
		--	Q => x_prime_reg);		
			
	store_tweak: entity work.regn(behavioral)
		 GENERIC MAP(N=>128)
		 PORT MAP(
		   CLK => CLK,
			--RESET => RESET,
			EN => EN,
			D => Next_T,
			Q => T);

   inc_sigma: entity work.regn_en(behavioral)
		GENERIC MAP(n => 4)
		PORT MAP(
			d => next_sigma,
			clk => clk,
			--reset => rst,
			init => '0',
			en => en,
			q => sigma
			);

T_out <= T;

X_OUT <= X; 

-- Reverse endian matches input alignment with Ek and Dk operations

	reverse_X_in: entity work.Reverse_Endian(dataflow)
	    PORT MAP(
		    A => X_IN_rev,
			 B => X_IN);
			 
	reverse_X_out: entity work.Reverse_Endian(dataflow)
	    PORT MAP(
		    A => X_OUT,
			 B => X_OUT_rev);

   reverse_initial_tweak: entity work.Reverse_Endian(dataflow)
	    PORT MAP(
		    A => T0_rev,
			 B => T0);

   reverse_K: entity work.Reverse_Endian(dataflow)
	    PORT MAP(
		    A => K_rev,
			 B => K);

 
END structural;
