-------------------------------------------------------------------------------
--! @file       Ek_step_bidir.vhd
--! @author     William Diehl
--! @brief      SCREAM Ek_step_bidir
--! @date       29 Jun 2016
-------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;

ENTITY Ek_Step_bidir IS
	PORT(
	 
	 X_IN : IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 X_OUT : OUT STD_LOGIC_VECTOR(127 DOWNTO 0);
	 SIGMA  : IN STD_LOGIC_VECTOR(3 DOWNTO 0);
	 T: IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 Next_T : OUT STD_LOGIC_VECTOR(127 DOWNTO 0);
	 K: IN STD_LOGIC_VECTOR(127 DOWNTO 0);
	 inv : in std_logic); -- inv=0 encrypt; inv=1 decrypt

END Ek_Step_bidir;

ARCHITECTURE structural of Ek_Step_bidir IS

CONSTANT RHO0 : STD_LOGIC:= '0';
CONSTANT RHO1 : STD_LOGIC:= '1';
SIGNAL X0e, X1e, X2e, X0d, X1d, X2d : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL A0e, A1e, B0e, B1e : STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL A0d, A1d, B0d, B1d : STD_LOGIC_VECTOR(127 DOWNTO 0);

SIGNAL C0e, C1e : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL C0d, C1d : STD_LOGIC_VECTOR(15 DOWNTO 0);

SIGNAL RC0, RC1 : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL X_before_TK, TK, TSIGMA, TSIGMAe, TSIGMAd: STD_LOGIC_VECTOR(127 DOWNTO 0);
SIGNAL Txor : std_logic_vector(63 downto 0);

BEGIN


-- Global 

compute_RC_0: entity work.RC_LUT(dataflow)
	  PORT MAP(
		sigma => sigma,
		rho => RHO0,
		RC => RC0
		);

compute_RC_1: entity work.RC_LUT(dataflow)
	  PORT MAP(
		sigma => sigma,
		rho => RHO1,
		RC => RC1
		);

Txor <= T(127 DOWNTO 64) XOR T(63 DOWNTO 0);

-- Encrypt sequence

--- Round 0

	compute_S_Box_0: entity work.Sbox(structural)
		PORT MAP(
		X_IN => X_IN,
		X_OUT => A0e);	

		
     C0e <= A0e(127 DOWNTO 112) XOR RC0;
-- RC addition to Most significant word modulo 65536

   B0e<= C0e & A0e(111 DOWNTO 0);
-- Add Round Constant to Most significant word of status register

    compute_L_Box_0: entity work.L_Box_128_alt(structural)
	PORT MAP(
		
		x => B0e,
	 	w => X0e);

--- End Round  0

-- Round 1

	compute_S_Box_1: entity work.Sbox(structural)
		PORT MAP(
		X_IN => X0e,
		X_OUT => A1e);	
	
		
     C1e <= A1e(127 DOWNTO 112) XOR RC1;
-- RC addition to Most significant word modulo 65536

   B1e<= C1e & A1e(111 DOWNTO 0);
-- Add Round Constant to Most significant word of status register

-- L_Box_128 is 8-bit LBoxes
-- L_Box_128_alt is 16-bit LBoxes

    compute_L_Box_1: entity work.L_Box_128_alt(structural)
	PORT MAP(
		
	      x => B1e,
	      w => X1e);

TSIGMAe <= Txor & T(127 DOWNTO 64);

-- End Round 1

-- Decrypt sequence

-- Round 1

-- Inv_L_Box_128 is 8-bit Inv LBox
-- Inv_L_Box_128_alt is 16-bit Inv LBox

   compute_Inv_L_Box_1: entity work.Inv_L_Box_128_alt(structural)
	PORT MAP(
	      x => X_IN,
	      w => A1d);
	

     C1d <= A1d(127 DOWNTO 112) XOR RC1;
-- RC addition to most signficant word modulo 65536

     B1d <= C1d & A1d(111 DOWNTO 0);

-- Add Round Constant to Most significant word of status register

-- Inv_S_Box_alt1 is bitslice
-- Inv_S_Box_alt is LUT

	compute_Inv_S_Box_1: entity work.InvSbox(structural)
	   PORT MAP(
		X_IN => B1d,
		X_OUT => X1d);

-- End Round 1
		
-- Round 0

   compute_Inv_L_Box_0: entity work.Inv_L_Box_128_alt(structural)
	PORT MAP(

	      x => X1d,
	      w => A0d);
		

     C0d <= A0d(127 DOWNTO 112) XOR RC0;
-- RC addition to most significant word modulo 65536

     B0d <= C0d & A0d(111 DOWNTO 0);

-- Add Round Constant to Most significant word of status register

-- Inv_S_Box_alt1 is bitslice
-- Inv_S_Box_alt is LUT

	compute_Inv_S_Box_0: entity work.InvSbox(structural)
	   PORT MAP(
		X_IN => B0d,
		X_OUT => X0d);

TSIGMAd <= T(63 DOWNTO 0) & Txor;

-- End Round 0


-- compute next tweak

X_before_TK <= x1e when (inv='0') else x0d;
TSIGMA <= TSIGMAe when (inv='0') else TSIGMAd;

-- Compute next tweak

TK <= TSIGMA XOR K;
-- Compute Step Tweak Key

X_OUT <= X_before_TK XOR TK;
-- Input XOR with Tweak Key

Next_T <= TSIGMA;

END structural;
